/******************************************************************************/
/* cicvis1 - network traffic visulaisation                                    */
/* copyright (c) 2004 jamie twycross, jpt@cs.nott.ac.uk                       */
/* released under the gnu general public license v2                           */
/******************************************************************************/

static const char RCSid[] = "$Id: cicvis1.c,v 1.4 2004/03/02 22:43:45 jpt Exp $";
const char *RCScruft7 = RCSid;

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <pcap.h>
#include <errno.h>
#include <math.h>

#define INCLUDE_CICHLID_UTILS
#include <cichlid_server.h>

#define NUM_VERTICES	5
#define NUM_EDGES	4
#define NUM_LINKS	6
#define SPHERE_SIZE	1.2
#define CUBE_SIZE	0.7
#define PYRAMID_SIZE	0.7
#define EDGE_SIZE	0.1
#define VERTEX_COLOR_1	{0, 0, 255}
#define VERTEX_COLOR_2	{255, 0, 0}
#define VERTEX_COLOR_3	{0, 255, 0}
#define EDGE_COLOR_1	{0, 255, 0}
#define LABEL_COLOR	255
#define SPACE_SIZE	{20.0, 20.0, 20.0}

static struct DataSet *dset;
static const float rspace[3] = SPACE_SIZE;

static float edge_load[NUM_EDGES];

static pthread_t writer;
static pthread_mutex_t matrix_lock;
static char *device = {"eth0"};

static struct VtxInfo vertices[NUM_VERTICES] =
{
	{ {0.0, 0.0, 0.0}, VD_PYRAMID_3, PYRAMID_SIZE, VERTEX_COLOR_3 }, // cat
	{ {0.0, 5.0, -2.5}, VD_SPHERE, SPHERE_SIZE, VERTEX_COLOR_2 },  // ringo
	{ {4.33, -2.5, -2.5}, VD_SPHERE, SPHERE_SIZE, VERTEX_COLOR_2 },  // george
	{ {-4.33, -2.5, -2.5}, VD_SPHERE, SPHERE_SIZE, VERTEX_COLOR_2 },  // wilde
	{ {0.0, 0.0, 1.5}, VD_CUBE, CUBE_SIZE, VERTEX_COLOR_1 }  // jordan
};

static struct EdgeInfo edges[NUM_EDGES] =
{
	{ {0, 1}, ED_NONE, ER_CYLINDER, EDGE_SIZE, EDGE_COLOR_1 },  // cat - ringo
	{ {0, 2}, ED_NONE, ER_CYLINDER, EDGE_SIZE, EDGE_COLOR_1 },  // cat - george
	{ {0, 3}, ED_NONE, ER_CYLINDER, EDGE_SIZE, EDGE_COLOR_1 },  // cat - wilde
	{ {0, 4}, ED_NONE, ER_CYLINDER, EDGE_SIZE, EDGE_COLOR_1 }  // cat - jordan
};

static char *vertex_labels[NUM_VERTICES] =
{
	"cat",
	"ringo",
	"george",
	"wilde",
	"jordan"
};

static unsigned char edge_info[NUM_LINKS][2][6] = 
{
	{{0x00,0x0C,0x41,0x27,0x0B,0xE1},   // jprdan - ringo
	{0x00,0x50,0x04,0x4A,0x2C,0x41}},
	{{0x00,0x0C,0x41,0x27,0x0B,0xE1},   // jprdan - george
	{0x00,0x0A,0x5E,0x20,0x4A,0x82}}, 
	{{0x00,0x0C,0x41,0x27,0x0B,0xE1},   // jordan - wilde
	{0x00,0x50,0x04,0x4A,0x2C,0x3C}}, 
	{{0x00,0x50,0x04,0x4A,0x2C,0x41},   // ringo - george
	{0x00,0x0A,0x5E,0x20,0x4A,0x82}}, 
	{{0x00,0x0A,0x5E,0x20,0x4A,0x82},   // george - wilde
	{0x00,0x50,0x04,0x4A,0x2C,0x3C}},
	{{0x00,0x50,0x04,0x4A,0x2C,0x3C},   // wilde - ringo
	{0x00,0x50,0x04,0x4A,0x2C,0x41}} 
};

static inline short int compare_mac(unsigned char *a, unsigned char *b)
{
	register unsigned int i;

	for(i = 0; i < 6; i++)
		if(a[i] != b[i])
			return 0;
	return 1;
}

static inline void print_mac(unsigned char *a)
{
	register unsigned int i;

	printf("%2X", a[0]);
	for(i = 1; i < 6; i++)
		printf(":%2X", a[i]);
	printf("\n");
}

static void packet_handler(u_char *dummy, const struct pcap_pkthdr *pkthdr, \
		const u_char *packet)
{
	register unsigned int link_num;
	register float load;
	unsigned char *shaddr, *dhaddr;

	if(pkthdr->caplen < 12)
		return;

	dhaddr = (unsigned char *) packet;
	shaddr = (unsigned char *) (packet + 6);
	for(link_num = 0; link_num < NUM_LINKS; link_num++)
	{
		if((compare_mac(edge_info[link_num][0], shaddr) && \
				compare_mac(edge_info[link_num][1], dhaddr)) ||
				(compare_mac(edge_info[link_num][1], shaddr) && \
				compare_mac(edge_info[link_num][0], dhaddr)))
			break;
	}
	if(link_num == NUM_LINKS)
		return;

	load = (float) pkthdr->len;
	pthread_mutex_lock(&matrix_lock);
	switch(link_num)
	{
		case 0:
			edge_load[0] += load;
			edge_load[3] += load;
			break;
			
		case 1:
			edge_load[1] += load;
			edge_load[3] += load;
			break;
			
		case 2:
			edge_load[2] += load;
			edge_load[3] += load;
			break;
			
		case 3:
			edge_load[0] += load;
			edge_load[1] += load;
			break;
			
		case 4:
			edge_load[1] += load;
			edge_load[2] += load;
			break;
			
		case 5:
			edge_load[0] += load;
			edge_load[2] += load;
			break;
			
		default:
			break;	
	}
	pthread_mutex_unlock(&matrix_lock);
}

void packet_sniffer(void)
{
	pcap_t *handle;
	bpf_u_int32 net, mask;
	char errbuf[PCAP_ERRBUF_SIZE];

	/* setup sniffing session */
	if(pcap_lookupnet(device, &net, &mask, errbuf))
		if(errno != 99)  /* allow no IP on iterface */
		{
			fprintf(stderr, "%s\n", errbuf);
			exit(EXIT_FAILURE);
		}
	handle = pcap_open_live(device, BUFSIZ, 0, -1, errbuf);
	if(!handle)
	{
		fprintf(stderr, "%s\n", errbuf);
		exit(EXIT_FAILURE);
	}

	/* handle the packets */
	pcap_loop(handle, -1, packet_handler, NULL);
}

static void data_init(void)
{
	unsigned int i;

	veset_set_vtcs(dset, 0, &vertices[0], NUM_VERTICES);
	veset_set_edges(dset, 0, &edges[0], NUM_EDGES);

	for(i = 0; i < NUM_VERTICES; i++)
		veset_set_vlabel(dset, i, vertex_labels[i], NULL);

	for(i = 0; i < NUM_EDGES; i++)
		edge_load[i] = 0.0;
}

static inline float scale_load(float a)
{
	return((1.0 / (1.0 + exp(-1.0 * a / 2.0))) - 0.5);
}

static void data_calc(void)
{
	register unsigned int i;

	pthread_mutex_lock(&matrix_lock);
	for(i = 0; i < NUM_EDGES; i++)
	{
		edges[i].size = EDGE_SIZE + scale_load(edge_load[i]);
		edge_load[i] = 0.0;
	}
	pthread_mutex_unlock(&matrix_lock);

	veset_set_vtcs(dset, 0, &vertices[0], NUM_VERTICES);
	veset_set_edges(dset, 0, &edges[0], NUM_EDGES);
}

int main(int argc, char **argv)
{
	dsInit(&argc, argv, TB_FALSE);
	dset = veset_new("cicvis1", 0, NUM_VERTICES, NUM_EDGES, rspace, CS_RECTANGULAR);
	data_init();
	dsSetDataSet(dset);
	dsDataFunc(data_calc);

	pthread_mutex_init(&matrix_lock, NULL);
	pthread_create(&writer, NULL, (void *) &packet_sniffer, NULL);

	dsMainLoop();

	return EXIT_SUCCESS;
}
